#include <time.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "ot.h"
#include "hyperorient.h"



#define LEFT_CUT 0
#define RIGHT_CUT 1
#define TIMING_INTERVAL 100000


or_tree * hyp_OrientWorstRecursive(//INPUTS:
			      dt_node * dt_root,           //root of dt
			      dt_node * dt_proto,          //node in dt corresponding to this node of or tree
			      hyp_LCAinfo ** lca_info,
			      hypergraph * g, inverse_hypergraph * h,
 			      //OUTPUTS:
			      or_tree ** solution,
			      int * local_cost,     
			      int * left_cut,
			      int * right_cut
#ifdef PROGRESS_METER
			      //info to track progress:
			      ,int ot_size
#endif
			      )
{
  int child;
  int winner;
  int cost[2];
  int possible_left_cut[2][2];
  int possible_right_cut[2][2];
  int possible_local_cost[2][2];
  int b;


#ifdef PROGRESS_METER
  static int depth = -1;
  static int ot_counter = 0;
  static char path[32] = {0};
  static int timing_start;
  static int time_left;
  static int min_left;
  static int hrs_left;
  static int sec_left;
#endif

  //Holds the two possible orientation trees
  or_tree * possible_solution[2];
  

#ifdef PROGRESS_METER
  int i,elapsed,timing_end;
  char spaces[20];
  for(i=0;i<20-depth;i++)
    spaces[i] = ' ';
  spaces[i] = 0;

#ifdef TIMER

  //  for(i=0; i<100000000;i++);

  //take a timing reading
  if(ot_counter % TIMING_INTERVAL == 0)
    {
      timing_end = time(NULL);
      elapsed = timing_end - timing_start;
      timing_start = time(NULL);
    
      time_left = (int)(((float)(ot_size - ot_counter)) / TIMING_INTERVAL) * elapsed;
      sec_left = time_left % 60;
      min_left = (time_left / 60) % 60;
      hrs_left = time_left / (60 * 60);
      
    }
#endif

  //Track progression of the algorithm
  
  printf("%s,%s %i / %i = %f",path, spaces, ++ot_counter, ot_size, ((float)ot_counter) / ot_size);

#ifdef TIMER
  printf("Remain(H:M:S): %i:%i:%i",hrs_left,min_left,sec_left);
#endif
  printf("\n");

  depth += 1;

#endif

  //If this is a leaf of the dt, then we need to do something completely different:
  if(dt_proto->v != NOT_LEAF)
    {
      //The orientation is irrelevant, so just return NULL
      (*solution = NULL);

      (*left_cut) = 0;
      (*right_cut) = 0;

      //Compute the outer cuts
      hyp_OuterCutsOfSingleVertex(//INPUTS:
				  dt_proto->v, g, h, lca_info,
#ifdef DEBUG
				  dt_root,
#endif

				  //OUTPUTS:
				  left_cut, right_cut);
#ifdef DEBUG
      printf("Left Cut: %i, Right Cut: %i\n",*left_cut,*right_cut); 
#endif

      (*local_cost) = (*left_cut);

#ifdef PROGRESS_METER
      path[depth] = 0;
      depth -= 1;
#endif

      return (*solution);

    }

  //allocate memory for orientation trees
  possible_solution[0] = New(or_tree);
  possible_solution[1] = New(or_tree);

#ifdef PROGRESS_METER
  path[depth] = '0';
#endif

  //for each possible orientation b
  for (b=0; b<=1; b++)
    {
      //orientation of root doesn't matter, so set it to 1, and skip 0
      //(Cuts running time in half)
      if(dt_proto == dt_root) b=1;


      //Set this orientation into dt for children's knowledge
      dt_proto->orientation = b;
      
      //   recursively find best orientation, local cost, and outer cuts
      //   of left and right children, given orientation b
      for(child=0; child<=1; child++)
	{
	  hyp_OrientWorstRecursive(//INPUTS
			      dt_root, 
			      dt_proto->child[child],
			      lca_info,
			      g,h,
			      //OUTPUTS
			      &(possible_solution[b]->child[child]),
			      &(possible_local_cost[b][child]),
			      &(possible_left_cut[b][child]),
			      &(possible_right_cut[b][child])
#ifdef PROGRESS_METER
			      ,ot_size
#endif
			      );
 
#ifdef PROGRESS_METER
	  ++path[depth];
	  //	  path[depth] = (path[depth] - '0') % 4 + '0';
#endif
	  
	}
        
      //   Compute local cost of this orientation based on values returned
      //   by two children.
      
      //The local cost depends on the problem being solved (Cut-width or Linear Arrangement)
      cost[b] = LocalCostOnTheFly( possible_local_cost[b][b],
				       possible_local_cost[b][1-b],
				       possible_right_cut[b][b],
				       possible_left_cut[b][1-b],
				       dt_proto->inner_cut,
				       dt_proto->child[b]->subtree_size,
				       dt_proto->child[1-b]->subtree_size);
    }
  
  //determine winner, set orientation;
  if(dt_proto == dt_root)
    winner = 1;
  else
    winner = (cost[0] <= cost[1]);
  
  possible_solution[winner]->orientation = winner;

  //Take better one as return value of local cost.
  (*local_cost) = cost[winner];
  
  //compute left and right cuts, return them.
  (*left_cut) = 
    possible_left_cut[winner][0] 
    + possible_left_cut[winner][1] 
    - dt_proto->inner_cut;

  (*right_cut) = 
    possible_right_cut[winner][0] 
    + possible_right_cut[winner][1] 
    - dt_proto->inner_cut;

  //reset this orientation for posterity
  dt_proto->orientation = -1;
  
  //Delete the or_tree of the loser
  if(dt_proto != dt_root)
    FreeOrTree(possible_solution[1-winner]);

  //return siolution of winner
  (*solution) = possible_solution[winner];
  
  
#ifdef PROGRESS_METER
  path[depth] = NULL;
  depth -= 1;
#endif

  //return local cost as ret value
  return (*solution);

}


or_tree * hyp_OrientWorstOnTheFly(hypergraph * g, dt_node * dt)
{

  hyp_LCAinfo ** lca_info = hyp_PreProcessLCA(g, dt);
  inverse_hypergraph * h = InverseHypergraph(g);

#ifdef PROGRESS_METER
  int ot_size = SizeOfOt(dt, 0);
#endif

  or_tree * ret;
  or_tree ** solution = New(or_tree*);
  int * local_cost = New(int); 
  int * left_cut = New(int);
  int * right_cut = New(int);
  
  
  ret = hyp_OrientWorstRecursive(//INPUTS:
			    dt,           //root of dt
			    dt,          //node in dt corresponding to this node of ot
			    lca_info,
			    g,h,
			    //OUTPUTS:
			    solution,
			    local_cost,     
			    left_cut,
			    right_cut
#ifdef PROGRESS_METER
			    //info for progress:
			    ,ot_size
#endif
			    );
  

#ifdef DEBUG3
  printf("Local Cost At Root: %i\n",*local_cost);
#endif

  free(local_cost);
  free(left_cut);
  free(right_cut);
  
  return ret;

}


or_tree * CreateZeroOrientation(dt_node * dt)
{ 

  or_tree * ret;
  
  if(dt->v != NOT_LEAF)
    return NULL;

  ret = New(or_tree);
  
  ret->orientation = 0;
  ret->child[0] = CreateZeroOrientation(dt->child[0]);
  ret->child[1] = CreateZeroOrientation(dt->child[1]);
  
  return ret;
}


or_tree * CreateRandomOrientation(dt_node * dt)
{
  or_tree * ret;
  
  if(dt->v != NOT_LEAF)
    return NULL;

  ret = New(or_tree);
  
  ret->orientation = RandomBit();
  ret->child[0] = CreateZeroOrientation(dt->child[0]);
  ret->child[1] = CreateZeroOrientation(dt->child[1]);
  
  return ret;



}
